""" Helpers for developping quick test scripts

Creation motivated by fear of filling disc space during long-running stress
tests

.. codeauthor:: Christian Herdtweck, christian.herdtweck@intra2net.com
"""

from __future__ import print_function

import contextlib
from threading import Thread
from time import sleep
from file_helpers import get_filesystem_fill_states
from datetime import datetime as dt
from buffers import LogarithmicBuffer
from itertools import tee
from warnings import warn
from sys import exit, stderr
from os import _exit as bad_exit_function_that_should_not_be_used

try:
    from warnings import ResourceWarning
    WARN_BASE_CLASS = ResourceWarning
except ImportError:
    # only added in python 3.2
    WARN_BASE_CLASS = UserWarning


class DiscFullPreventionError(Exception):
    """ Exception raised when disc space gets critical

    issued from function called by :py:class:`DiscFillCheckerThread`
    usually only happens after issuing a :py:class:`DiscFullPreventionWarning`
    """
    def __init__(self, state, time_estim):
        super(DiscFullPreventionError, self).__init__(
            'Interrupting test to avoid full disc ({0}, full in {1}s)'
            .format(state, "?" if time_estim==None else time_estim))
        self.state = state
        self.time_estim = time_estim


class DiscFullPreventionWarning(WARN_BASE_CLASS):
    """ Exception raised when disc space approaches problematic state

    issued from function called by :py:class:`DiscFillCheckerThread`
    If disc fills up further, will be followed by a
    :py:class:`DiscFullPreventionError` or sys.exit
    """
    def __init__(self, state, time_estim):
        super(DiscFullPreventionWarning, self).__init__(
              'Disc nearly full! Might soon abort test ({0}, '
              'full in {1}s)'
              .format(state, "?" if time_estim==None else time_estim))
        self.state = state
        self.time_estim = time_estim


class DiscFillCheckerThread(Thread):
    """ a thread that checks disc fill in regular intervals """

    def __init__(self, interval=10, decision_function=None):
        """ creates a DiscFillCheckerThread

        :param int interval: time in seconds between checks of disc usage
        :param decision_function: function that decides when to issue a warning
                                 or an error. See
                                 :py:func:`default_disc_full_decision_function`
                                 as an example. Function should raise
                                 :py:class:`DiscFullPreventionWarning` and
                                 :py:class:`DiscFullPreventionError`
        """
        super(DiscFillCheckerThread, self).__init__(name='discFillChck')

        self.daemon = True   # works since python 2.6

        # set variables
        self.interval = interval
        if decision_function is None:
            self.decision_function = default_disc_full_decision_function
        else:
            # need to check warn_func somehow (using introspection?)
            self.decision_function = decision_function

        # remember relevant disc stats at start
        self._bufs = {}
        for fs_state in get_filesystem_fill_states():
            self._internal_state_buffer(fs_state)

    def _internal_state_buffer(self, fs_state):
        """ update internal stats buffer, returns all estims

        internal helper called from __init__ and run
        """
        if fs_state.name == 'tmpfs':
            print('ignoring {0}'.format(fs_state))
            return []

        buf = None
        try:
            buf = self._bufs[fs_state.name]
        except KeyError:
            # new file system -- need to create a new buffer
            buf = LogarithmicBuffer(5)
            self._bufs[fs_state.name] =  buf

        buf.add((dt.now(), float(fs_state.available)))
        return buf.get_all()

    def run(self):
        while True:
            sleep(self.interval)
            fs_states = get_filesystem_fill_states()
            for fs_state in fs_states:
                # ignore tmpfs since is in memory or swap, so fill up is not
                # as bad
                if fs_state.name == 'tmpfs':
                    continue

                fill_states = self._internal_state_buffer(fs_state)

                # estimate time until full (i.e. when available == 0)
                times_until_empty = \
                    [calc_time_until_zero(old_state, new_state)
                     for old_state, new_state in pairwise(fill_states)]

                # call user-defined function to decide
                min_time = None
                if any(time > 0 for time in times_until_empty):
                    min_time = min(time for time in times_until_empty
                                   if time>0)
                avg_time = calc_time_until_zero(fill_states[0],
                                                 fill_states[-1])
                self.decision_function(fs_state,
                                       times_until_empty[-1],     # newest
                                       min_time,                  # smallest
                                       avg_time)                  # average


MEGABYTE = 2**20
MINUTE = 60

#: default thresholds for :py:func:`default_disc_full_decision_function`
DEFAULT_DISC_SIZE_WARN = 10 * MEGABYTE
DEFAULT_DISC_SIZE_ERR = MEGABYTE
DEFAULT_TIME_TO_FULL_WARN = 10 * MINUTE
DEFAULT_TIME_TO_FULL_ERR = MINUTE

#: levels at which to kill running tests: regular exception (subclass of
#: Exception), base exception (to avoid "catch(exception)") or sys.exit
KILL_EXCEPTION = "KILL_EXCEPTION"
KILL_SYS_EXIT = "KILL_SYS_EXIT"
KILL_REALLY_MEAN_EXIT = "REALLY_MEAN_EXIT"

def default_disc_full_decision_function(curr_state,
                                        estim_empty_newest,
                                        estim_empty_smallest,
                                        estim_empty_mean,
                                        size_warn=DEFAULT_DISC_SIZE_WARN,
                                        size_err=DEFAULT_DISC_SIZE_ERR,
                                        time_warn=DEFAULT_TIME_TO_FULL_WARN,
                                        time_err=DEFAULT_TIME_TO_FULL_ERR,
                                        kill_level=KILL_EXCEPTION):
    """ decide whether to warn or raise error because disc is getting full

    called by DiscFillCheckerThread regularly; params estim_empty_* are result
    of :py:func:`calc_time_until_zero`, so None means that disc state did not
    change.

    Issues a :py:class:`DiscFullPreventionWarning` when space gets problematic
    and raises an error or calls sys.exit when it is really critical.

    :param estim_empty_newest: the newest time estimate
    :param estim_empty_smallest: the smallest of all non-negative estimates
    :param estim_empty_mean: estim between earliest and newest estimate
    :param int size/err_warn/err: thresholds, default to constants like
                                  :py:data:`DEFAULT_DISC_SIZE_WARN`
    :param kill_level: how to kill the running application: with sys.exit or
                       by raising a "regular" Exception subclass (default).
                       Values should be one like :py:data:`KILL_EXCEPTION`
    """

    # err?
    raise_err = False
    if curr_state.available < size_err:
        raise_err = True
    if (estim_empty_smallest != None) and (estim_empty_smallest < time_err):
        raise_err = True

    # what kind of err?
    if raise_err:
        if kill_level == KILL_EXCEPTION:
            raise DiscFullPreventionError(curr_state, estim_empty_smallest)
        elif kill_level == KILL_REALLY_MEAN_EXIT:
            print('Disc fill below tolerance or disc full soon', file=stderr)
            print('{0}, full in {1} s'
                  .format(curr_state, estim_empty_smallest), file=stderr)
            print('Exiting now the brutal way', file=stderr)
            bad_exit_function_that_should_not_be_used(1)
        else:   # kill_level == SYS_EXIT or unknown
            print('Disc fill below tolerance or disc full soon:', file=stderr)
            print('{0}, full in {1} s'
                  .format(curr_state, estim_empty_smallest), file=stderr)
            print('Exiting now', file=stderr)
            exit("Disc nearly full!")

    # warn?
    if curr_state.available < size_warn:
        warn(DiscFullPreventionWarning(curr_state, estim_empty_smallest),
             stacklevel=3)
    if (estim_empty_smallest != None) and (estim_empty_smallest < time_warn):
        warn(DiscFullPreventionWarning(curr_state, estim_empty_smallest),
             stacklevel=3)


def calc_time_until_zero(old_state, new_state):
    """ calc time until value hits zero given (time1, value1), (time2, value2)

    returns time from now until empty in seconds

    delta_date = new_date - old_date
    delta_free = new_free - old_free
    free(t) = new_free + (t-new_t) * delta_free / delta_date
            = 0
    <==>  -new_free = (t-new_t) * delta_free / delta_date
    <==>  -new_free * delta_date / date_free = t - new_t
    <==>  diff_t := new_t - t = new_free * delta_date / delta_free

    compare to now:
    diff_to_now = t + (-new_t + new_t) - now = (t - new_t) + (new_t - now)
                = -diff_t - (now-new_t) = (new_t-now) - diff_t

    To avoid zero-division, returns None if delta_free == 0
    """
    old_date, old_free = old_state
    new_date, new_free = new_state
    if new_free == old_free:
        return None            # avoid zero-division
    time_diff = new_free * (new_date - old_date).total_seconds() \
                         / (new_free - old_free)

    # compare to now
    return (new_date - dt.now()).total_seconds() - time_diff


def pairwise(iterable):
    """ s -> (s0,s1), (s1,s2), (s2, s3), ...

    taken from itertool recipes in python api doc of itertools
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


@contextlib.contextmanager
def disc_fill_checked(*args, **kwargs):
    """ run test function while separate thread watches disc space

    all args are forwarded to :py:class:`DiscFillCheckerThread` constructor
    """

    # todo: add args and forward
    DiscFillCheckerThread(*args, **kwargs).start()

    try:
        yield
    finally:
        pass # wait for other to stop or so? or stop it?
